@@ -49,7 +49,7 @@ static struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.audioadapter");
 
 #define DEFAULT_ALIGN	16
 
-#define MAX_PORTS	SPA_AUDIO_MAX_CHANNELS
+#define MAX_PORTS	(SPA_AUDIO_MAX_CHANNELS+1)
 
 /** \cond */
 
@@ -257,6 +257,9 @@ static void emit_node_info(struct impl *this, bool full)
 	uint32_t i;
 	uint64_t old = full ? this->info.change_mask : 0;
 
+	spa_log_debug(this->log, "%p: info full:%d change:%08"PRIx64,
+			this, full, this->info.change_mask);
+
 	if (full)
 		this->info.change_mask = this->info_all;
 	if (this->info.change_mask) {
@@ -265,6 +268,8 @@ static void emit_node_info(struct impl *this, bool full)
 				if (this->params[i].user > 0) {
 					this->params[i].flags ^= SPA_PARAM_INFO_SERIAL;
 					this->params[i].user = 0;
+					spa_log_debug(this->log, "param %d flags:%08x",
+							i, this->params[i].flags);
 				}
 			}
 		}
@@ -277,38 +282,45 @@ static int debug_params(struct impl *this, struct spa_node *node,
                 enum spa_direction direction, uint32_t port_id, uint32_t id, struct spa_pod *filter,
 		const char *debug, int err)
 {
-        struct spa_pod_builder b = { 0 };
-        uint8_t buffer[4096];
-        uint32_t state;
-        struct spa_pod *param;
-        int res;
+	struct spa_pod_builder b = { 0 };
+	uint8_t buffer[4096];
+	uint32_t state;
+	struct spa_pod *param;
+	int res, count = 0;
 
-        spa_log_error(this->log, "params %s: %d:%d (%s) %s",
+	spa_log_error(this->log, "params %s: %d:%d (%s) %s",
 			spa_debug_type_find_name(spa_type_param, id),
-			direction, port_id, debug, spa_strerror(err));
+			direction, port_id, debug, err ? spa_strerror(err) : "no matching params");
 	if (err == -EBUSY)
 		return 0;
 
-        state = 0;
-        while (true) {
-                spa_pod_builder_init(&b, buffer, sizeof(buffer));
-                res = spa_node_port_enum_params_sync(node,
-                                       direction, port_id,
-                                       id, &state,
-                                       NULL, &param, &b);
-                if (res != 1) {
+	if (filter) {
+		spa_log_error(this->log, "with this filter:");
+		spa_debug_pod(2, NULL, filter);
+	} else {
+		spa_log_error(this->log, "there was no filter");
+	}
+
+	state = 0;
+	while (true) {
+		spa_pod_builder_init(&b, buffer, sizeof(buffer));
+		res = spa_node_port_enum_params_sync(node,
+					direction, port_id,
+					id, &state,
+					NULL, &param, &b);
+		if (res != 1) {
 			if (res < 0)
 				spa_log_error(this->log, "  error: %s", spa_strerror(res));
-                        break;
+			break;
 		}
-                spa_debug_pod(2, NULL, param);
-        }
+		spa_log_error(this->log, "unmatched %s %d:", debug, count);
+		spa_debug_pod(2, NULL, param);
+		count++;
+	}
+	if (count == 0)
+		spa_log_error(this->log, "could not get any %s", debug);
 
-        spa_log_error(this->log, "failed filter:");
-        if (filter)
-                spa_debug_pod(2, NULL, filter);
-
-        return 0;
+	return 0;
 }
 
 static int negotiate_buffers(struct impl *this)
@@ -430,6 +442,29 @@ static int configure_format(struct impl *this, uint32_t flags, const struct spa_
 	if (format && spa_log_level_enabled(this->log, SPA_LOG_LEVEL_DEBUG))
 		spa_debug_format(0, NULL, format);
 
+	if ((res = spa_node_port_set_param(this->follower,
+					   this->direction, 0,
+					   SPA_PARAM_Format, flags,
+					   format)) < 0)
+			return res;
+	if (res > 0) {
+		uint8_t buffer[4096];
+		struct spa_pod_builder b = { 0 };
+		uint32_t state = 0;
+		struct spa_pod *fmt;
+
+		/* format was changed to nearest compatible format */
+		spa_pod_builder_init(&b, buffer, sizeof(buffer));
+
+		if ((res = spa_node_port_enum_params_sync(this->follower,
+					this->direction, 0,
+					SPA_PARAM_Format, &state,
+					NULL, &fmt, &b)) != 1)
+			return -EIO;
+
+		format = fmt;
+	}
+
 	if (this->target != this->follower && this->convert) {
 		if ((res = spa_node_port_set_param(this->convert,
 					   SPA_DIRECTION_REVERSE(this->direction), 0,
@@ -438,12 +473,6 @@ static int configure_format(struct impl *this, uint32_t flags, const struct spa_
 				return res;
 	}
 
-	if ((res = spa_node_port_set_param(this->follower,
-					   this->direction, 0,
-					   SPA_PARAM_Format, flags,
-					   format)) < 0)
-			return res;
-
 	this->have_format = format != NULL;
 	if (format == NULL) {
 		this->n_buffers = 0;
@@ -482,7 +511,7 @@ static int reconfigure_mode(struct impl *this, bool passthrough,
 	int res = 0;
 	struct spa_hook l;
 
-	spa_log_debug(this->log, "%p: passthrough mode %d", this, passthrough);
+	spa_log_info(this->log, "%p: passthrough mode %d", this, passthrough);
 
 	if (this->passthrough != passthrough) {
 		if (passthrough) {
@@ -501,7 +530,7 @@ static int reconfigure_mode(struct impl *this, bool passthrough,
 	/* set new target */
 	this->target = passthrough ? this->follower : this->convert;
 
-	if ((res = configure_format(this, 0, format)) < 0)
+	if ((res = configure_format(this, SPA_NODE_PARAM_FLAG_NEAREST, format)) < 0)
 		return res;
 
 	if (this->passthrough != passthrough) {
@@ -542,7 +571,6 @@ static int format_audio_raw_parse_opt(const struct spa_pod *format, struct spa_a
 	res = spa_pod_parse_object(format,
 			SPA_TYPE_OBJECT_Format, NULL,
 			SPA_FORMAT_AUDIO_format, SPA_POD_OPT_Id(&info->format),
-			SPA_FORMAT_AUDIO_rate, SPA_POD_OPT_Int(&info->rate),
 			SPA_FORMAT_AUDIO_channels, SPA_POD_OPT_Int(&info->channels),
 			SPA_FORMAT_AUDIO_position, SPA_POD_OPT_Pod(&position));
 	if (position == NULL ||
@@ -584,7 +612,6 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 		enum spa_direction dir;
 		enum spa_param_port_config_mode mode;
 		struct spa_pod *format = NULL;
-		int monitor = false;
 
 		if (this->started) {
 			spa_log_error(this->log, "was started");
@@ -595,7 +622,6 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 				SPA_TYPE_OBJECT_ParamPortConfig, NULL,
 				SPA_PARAM_PORT_CONFIG_direction,	SPA_POD_Id(&dir),
 				SPA_PARAM_PORT_CONFIG_mode,		SPA_POD_Id(&mode),
-				SPA_PARAM_PORT_CONFIG_monitor,		SPA_POD_OPT_Bool(&monitor),
 				SPA_PARAM_PORT_CONFIG_format,		SPA_POD_OPT_Pod(&format)) < 0)
 			return -EINVAL;
 
@@ -766,7 +792,7 @@ static int negotiate_format(struct impl *this)
 
 	spa_pod_fixate(format);
 
-	res = configure_format(this, 0, format);
+	res = configure_format(this, SPA_NODE_PARAM_FLAG_NEAREST, format);
 
 done:
 	spa_node_send_command(this->follower,
@@ -833,11 +859,20 @@ static void convert_node_info(void *data, const struct spa_node_info *info)
 	struct impl *this = data;
 	uint32_t i;
 
+	spa_log_debug(this->log, "%p: info change:%08"PRIx64, this,
+			info->change_mask);
+
 	if (info->change_mask & SPA_NODE_CHANGE_MASK_PARAMS) {
 		for (i = 0; i < info->n_params; i++) {
 			uint32_t idx;
 
 			switch (info->params[i].id) {
+			case SPA_PARAM_EnumPortConfig:
+				idx = IDX_EnumPortConfig;
+				break;
+			case SPA_PARAM_PortConfig:
+				idx = IDX_PortConfig;
+				break;
 			case SPA_PARAM_PropInfo:
 				idx = IDX_PropInfo;
 				break;
@@ -857,8 +892,10 @@ static void convert_node_info(void *data, const struct spa_node_info *info)
 				(this->params[idx].flags & SPA_PARAM_INFO_SERIAL) |
 				(info->params[i].flags & SPA_PARAM_INFO_READWRITE);
 
-			if (!this->add_listener)
+			if (!this->add_listener) {
 				this->params[idx].user++;
+				spa_log_debug(this->log, "param %d changed", info->params[i].id);
+			}
 		}
 	}
 	emit_node_info(this, false);
@@ -877,7 +914,7 @@ static void convert_port_info(void *data,
 			port_id--;
 	}
 
-	spa_log_trace(this->log, "%p: port info %d:%d", this,
+	spa_log_debug(this->log, "%p: port info %d:%d", this,
 			direction, port_id);
 
 	if (this->target != this->follower)
@@ -907,6 +944,9 @@ static void follower_info(void *data, const struct spa_node_info *info)
 	struct impl *this = data;
 	uint32_t i;
 
+	spa_log_debug(this->log, "%p: info change:%08"PRIx64, this,
+			info->change_mask);
+
 	if (this->follower_removing)
 		return;
 
@@ -960,11 +1000,17 @@ static void follower_info(void *data, const struct spa_node_info *info)
 				(this->params[idx].flags & SPA_PARAM_INFO_SERIAL) |
 				(info->params[i].flags & SPA_PARAM_INFO_READWRITE);
 
-			if (!this->add_listener)
+			if (!this->add_listener) {
 				this->params[idx].user++;
+				spa_log_debug(this->log, "param %d changed", info->params[i].id);
+			}
 		}
 	}
 	emit_node_info(this, false);
+
+	spa_zero(this->info.props);
+	this->info.change_mask &= ~SPA_NODE_CHANGE_MASK_PROPS;
+
 }
 
 static int recalc_latency(struct impl *this, enum spa_direction direction, uint32_t port_id)
@@ -1045,12 +1091,15 @@ static void follower_port_info(void *data,
 
 			if (idx == IDX_Latency) {
 				res = recalc_latency(this, direction, port_id);
-				spa_log_debug(this->log, "latency: %d", res);
+				spa_log_debug(this->log, "latency: %d (%s)", res,
+						spa_strerror(res));
 			}
 
 			this->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
-			if (!this->add_listener)
+			if (!this->add_listener) {
 				this->params[idx].user++;
+				spa_log_debug(this->log, "param %d changed", info->params[i].id);
+			}
 		}
 	}
 	emit_node_info(this, false);
@@ -1070,11 +1119,29 @@ static void follower_result(void *data, int seq, int res, uint32_t type, const v
 	spa_node_emit_result(&this->hooks, seq, res, type, result);
 }
 
+static void follower_event(void *data, const struct spa_event *event)
+{
+	struct impl *this = data;
+
+	spa_log_trace(this->log, "%p: event %d", this, SPA_EVENT_TYPE(event));
+
+	switch (SPA_NODE_EVENT_ID(event)) {
+	case SPA_NODE_EVENT_Error:
+		/* Forward errors */
+		spa_node_emit_event(&this->hooks, event);
+		break;
+	default:
+		/* Ignore other events */
+		break;
+	}
+}
+
 static const struct spa_node_events follower_node_events = {
 	SPA_VERSION_NODE_EVENTS,
 	.info = follower_info,
 	.port_info = follower_port_info,
 	.result = follower_result,
+	.event = follower_event,
 };
 
 static int follower_ready(void *data, int status)
@@ -1086,8 +1153,21 @@ static int follower_ready(void *data, int status)
 	if (this->target != this->follower) {
 		this->driver = true;
 
-		if (this->direction == SPA_DIRECTION_OUTPUT)
-			status = spa_node_process(this->convert);
+		if (this->direction == SPA_DIRECTION_OUTPUT) {
+			int retry = 8;
+			while (retry--) {
+				status = spa_node_process(this->convert);
+				if (status & SPA_STATUS_HAVE_DATA)
+					break;
+
+				if (status & SPA_STATUS_NEED_DATA) {
+					status = spa_node_process(this->follower);
+					if (!(status & SPA_STATUS_HAVE_DATA))
+						break;
+				}
+			}
+
+		}
 	}
 
 	return spa_node_call_ready(&this->callbacks, status);
@@ -1314,7 +1394,7 @@ impl_node_port_reuse_buffer(void *object, uint32_t port_id, uint32_t buffer_id)
 static int impl_node_process(void *object)
 {
 	struct impl *this = object;
-	int status = 0, fstatus;
+	int status = 0, fstatus, retry = 8;
 
 	spa_log_trace_fp(this->log, "%p: process convert:%p driver:%d",
 			this, this->convert, this->driver);
@@ -1329,7 +1409,7 @@ static int impl_node_process(void *object)
 		/* an input node (sink).
 		 * First we run the converter to process the input for the follower
 		 * then if it produced data, we run the follower. */
-		while (true) {
+		while (retry--) {
 			status = this->convert ? spa_node_process(this->convert) : 0;
 			/* schedule the follower when the converter needed
 			 * a recycled buffer */
@@ -1358,7 +1438,7 @@ static int impl_node_process(void *object)
 		}
 	} else if (!this->driver) {
 		bool done = false;
-		while (true) {
+		while (retry--) {
 			/* output node (source). First run the converter to make
 			 * sure we push out any queued data. Then when it needs
 			 * more data, schedule the follower. */
