@@ -101,41 +101,23 @@ static void fill_properties(struct pw_context *context)
 	pw_properties_set(properties, PW_KEY_CORE_NAME, context->core->info.name);
 }
 
-static int try_load_conf(struct pw_context *this, const char *conf_prefix,
-		const char *conf_name, struct pw_properties *conf)
-{
-	int res;
-
-	if (conf_name == NULL)
-		return -EINVAL;
-	if (spa_streq(conf_name, "null"))
-		return 0;
-	if ((res = pw_conf_load_conf(conf_prefix, conf_name, conf)) < 0) {
-		bool skip_prefix = conf_prefix == NULL || conf_name[0] == '/';
-		pw_log_warn("%p: can't load config %s%s%s: %s",
-				this,
-				skip_prefix ? "" : conf_prefix,
-				skip_prefix ? "" : "/",
-				conf_name, spa_strerror(res));
-	}
-	return res;
-}
-
 static int context_set_freewheel(struct pw_context *context, bool freewheel)
 {
 	struct spa_thread *thr;
-	int res;
+	int res = 0;
 
 	if ((thr = pw_data_loop_get_thread(context->data_loop_impl)) == NULL)
 		return -EIO;
 
 	if (freewheel) {
 		pw_log_info("%p: enter freewheel", context);
-		res = pw_thread_utils_drop_rt(thr);
+		if (context->thread_utils)
+			res = spa_thread_utils_drop_rt(context->thread_utils, thr);
 	} else {
 		pw_log_info("%p: exit freewheel", context);
-		// Use the priority as configured within the realtime module
-		res = pw_thread_utils_acquire_rt(thr, -1);
+		/* Use the priority as configured within the realtime module */
+		if (context->thread_utils)
+			res = spa_thread_utils_acquire_rt(context->thread_utils, thr, -1);
 	}
 	if (res < 0)
 		pw_log_info("%p: freewheel error:%s", context, spa_strerror(res));
@@ -209,7 +191,7 @@ struct pw_context *pw_context_new(struct pw_loop *main_loop,
 {
 	struct impl *impl;
 	struct pw_context *this;
-	const char *lib, *str, *conf_prefix, *conf_name;
+	const char *lib, *str;
 	void *dbus_iface = NULL;
 	uint32_t n_support;
 	struct pw_properties *pr, *conf;
@@ -268,23 +250,8 @@ struct pw_context *pw_context_new(struct pw_loop *main_loop,
 		goto error_free;
 	}
 	this->conf = conf;
-
-	conf_prefix = getenv("PIPEWIRE_CONFIG_PREFIX");
-	if (conf_prefix == NULL)
-		conf_prefix = pw_properties_get(properties, PW_KEY_CONFIG_PREFIX);
-
-	conf_name = getenv("PIPEWIRE_CONFIG_NAME");
-	if (try_load_conf(this, conf_prefix, conf_name, conf) < 0) {
-		conf_name = pw_properties_get(properties, PW_KEY_CONFIG_NAME);
-		if (try_load_conf(this, conf_prefix, conf_name, conf) < 0) {
-			conf_name = "client.conf";
-			if ((res = try_load_conf(this, conf_prefix, conf_name, conf)) < 0) {
-				pw_log_error("%p: can't load config %s: %s",
-					this, conf_name, spa_strerror(res));
-				goto error_free;
-			}
-		}
-	}
+	if ((res = pw_conf_load_conf_for_context (properties, conf)) < 0)
+		goto error_free;
 
 	n_support = pw_get_support(this->support, SPA_N_ELEMENTS(this->support) - 6);
 	cpu = spa_support_find(this->support, n_support, SPA_TYPE_INTERFACE_CPU);
@@ -411,8 +378,6 @@ struct pw_context *pw_context_new(struct pw_loop *main_loop,
 	pw_data_loop_invoke(this->data_loop_impl,
 			do_data_loop_setup, 0, NULL, 0, false, this);
 
-	context_set_freewheel(this, false);
-
 	pw_settings_expose(this);
 
 	pw_log_debug("%p: created", this);
@@ -464,7 +429,7 @@ void pw_context_destroy(struct pw_context *context)
 		pw_resource_destroy(resource);
 
 	if (context->data_loop_impl)
-		pw_data_loop_destroy(context->data_loop_impl);
+		pw_data_loop_stop(context->data_loop_impl);
 
 	spa_list_consume(module, &context->module_list, link)
 		pw_impl_module_destroy(module);
@@ -481,6 +446,9 @@ void pw_context_destroy(struct pw_context *context)
 	pw_log_debug("%p: free", context);
 	pw_context_emit_free(context);
 
+	if (context->data_loop_impl)
+		pw_data_loop_destroy(context->data_loop_impl);
+
 	if (context->pool)
 		pw_mempool_destroy(context->pool);
 
@@ -539,6 +507,12 @@ struct pw_loop *pw_context_get_main_loop(struct pw_context *context)
 	return context->main_loop;
 }
 
+SPA_EXPORT
+struct pw_data_loop *pw_context_get_data_loop(struct pw_context *context)
+{
+	return context->data_loop_impl;
+}
+
 SPA_EXPORT
 struct pw_work_queue *pw_context_get_work_queue(struct pw_context *context)
 {
@@ -618,98 +592,6 @@ struct pw_global *pw_context_find_global(struct pw_context *context, uint32_t id
 	return global;
 }
 
-/** Find a port to link with
- *
- * \param context a context
- * \param other_port a port to find a link with
- * \param id the id of a port or PW_ID_ANY
- * \param props extra properties
- * \param n_format_filters number of filters
- * \param format_filters array of format filters
- * \param[out] error an error when something is wrong
- * \return a port that can be used to link to \a otherport or NULL on error
- */
-struct pw_impl_port *pw_context_find_port(struct pw_context *context,
-				  struct pw_impl_port *other_port,
-				  uint32_t id,
-				  struct pw_properties *props,
-				  uint32_t n_format_filters,
-				  struct spa_pod **format_filters,
-				  char **error)
-{
-	struct pw_impl_port *best = NULL;
-	bool have_id;
-	struct pw_impl_node *n;
-
-	have_id = id != PW_ID_ANY;
-
-	pw_log_debug("%p: id:%u", context, id);
-
-	spa_list_for_each(n, &context->node_list, link) {
-		if (n->global == NULL)
-			continue;
-
-		if (other_port->node == n)
-			continue;
-
-		if (!global_can_read(context, n->global))
-			continue;
-
-		pw_log_debug("%p: node id:%d", context, n->global->id);
-
-		if (have_id) {
-			if (n->global->id == id) {
-				pw_log_debug("%p: id:%u matches node %p", context, id, n);
-
-				best =
-				    pw_impl_node_find_port(n,
-						pw_direction_reverse(other_port->direction),
-						PW_ID_ANY);
-				if (best)
-					break;
-			}
-		} else {
-			struct pw_impl_port *p, *pin, *pout;
-			uint8_t buf[4096];
-			struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buf, sizeof(buf));
-			struct spa_pod *dummy;
-
-			p = pw_impl_node_find_port(n,
-					pw_direction_reverse(other_port->direction),
-					PW_ID_ANY);
-			if (p == NULL)
-				continue;
-
-			if (p->direction == PW_DIRECTION_OUTPUT) {
-				pin = other_port;
-				pout = p;
-			} else {
-				pin = p;
-				pout = other_port;
-			}
-
-			if (pw_context_find_format(context,
-						pout,
-						pin,
-						props,
-						n_format_filters,
-						format_filters,
-						&dummy,
-						&b,
-						error) < 0) {
-				free(*error);
-				continue;
-			}
-			best = p;
-			break;
-		}
-	}
-	if (best == NULL) {
-		*error = spa_aprintf("No matching Node found");
-	}
-	return best;
-}
-
 SPA_PRINTF_FUNC(7, 8) int pw_context_debug_port_params(struct pw_context *this,
 		struct spa_node *node, enum spa_direction direction,
 		uint32_t port_id, uint32_t id, int err, const char *debug, ...)
@@ -912,27 +794,15 @@ static int ensure_state(struct pw_impl_node *node, bool running)
 	return pw_impl_node_set_state(node, state);
 }
 
-static int collect_nodes(struct pw_context *context, struct pw_impl_node *node)
+static int collect_nodes(struct pw_context *context, struct pw_impl_node *node, struct spa_list *collect)
 {
 	struct spa_list queue;
-	struct pw_impl_node *n, *t, *driver;
+	struct pw_impl_node *n, *t;
 	struct pw_impl_port *p;
 	struct pw_impl_link *l;
 
 	pw_log_debug("node %p: '%s'", node, node->name);
 
-	if (node->driver) {
-		driver = node;
-		spa_list_consume(t, &driver->follower_list, follower_link) {
-			spa_list_remove(&t->follower_link);
-			spa_list_init(&t->follower_link);
-		}
-	} else {
-		driver = node->driver_node;
-		if (driver == NULL)
-			return -EINVAL;
-	}
-
 	/* start with node in the queue */
 	spa_list_init(&queue);
 	spa_list_append(&queue, &node->sort_link);
@@ -941,10 +811,15 @@ static int collect_nodes(struct pw_context *context, struct pw_impl_node *node)
 	/* now follow all the links from the nodes in the queue
 	 * and add the peers to the queue. */
 	spa_list_consume(n, &queue, sort_link) {
+		pw_log_debug(" next node %p: '%s'", n, n->name);
+
 		spa_list_remove(&n->sort_link);
-		pw_impl_node_set_driver(n, driver);
+		spa_list_append(collect, &n->sort_link);
 		n->passive = true;
 
+		if (!n->active)
+			continue;
+
 		spa_list_for_each(p, &n->input_ports, link) {
 			spa_list_for_each(l, &p->links, input_link) {
 				t = l->output->node;
@@ -958,7 +833,7 @@ static int collect_nodes(struct pw_context *context, struct pw_impl_node *node)
 					continue;
 
 				if (!l->passive)
-					driver->passive = n->passive = false;
+					node->passive = n->passive = false;
 
 				if (!t->visited) {
 					t->visited = true;
@@ -979,7 +854,7 @@ static int collect_nodes(struct pw_context *context, struct pw_impl_node *node)
 					continue;
 
 				if (!l->passive)
-					driver->passive = n->passive = false;
+					node->passive = n->passive = false;
 
 				if (!t->visited) {
 					t->visited = true;
@@ -1005,6 +880,27 @@ static int collect_nodes(struct pw_context *context, struct pw_impl_node *node)
 	return 0;
 }
 
+static void move_to_driver(struct pw_context *context, struct spa_list *nodes,
+		struct pw_impl_node *driver)
+{
+	struct pw_impl_node *n;
+	pw_log_debug("driver: %p %s", driver, driver->name);
+	spa_list_consume(n, nodes, sort_link) {
+		spa_list_remove(&n->sort_link);
+		pw_log_debug(" follower: %p %s", n, n->name);
+		pw_impl_node_set_driver(n, driver);
+	}
+}
+static void remove_from_driver(struct pw_context *context, struct spa_list *nodes)
+{
+	struct pw_impl_node *n;
+	spa_list_consume(n, nodes, sort_link) {
+		spa_list_remove(&n->sort_link);
+		pw_impl_node_set_driver(n, NULL);
+		ensure_state(n, false);
+	}
+}
+
 static inline void get_quantums(struct pw_context *context, uint32_t *def,
 		uint32_t *min, uint32_t *max, uint32_t *limit, uint32_t *rate)
 {
@@ -1081,6 +977,24 @@ static bool rates_contains(uint32_t *rates, uint32_t n_rates, uint32_t rate)
 	return false;
 }
 
+/* here we evaluate the complete state of the graph.
+ *
+ * It roughly operates in 3 stages:
+ *
+ * 1. go over all drivers and collect the nodes that need to be scheduled with the
+ *    driver. This include all nodes that have an active link with the driver or
+ *    with a node already scheduled with the driver.
+ *
+ * 2. go over all nodes that are not assigned to a driver. The ones that require
+ *    a driver are moved to some random active driver found in step 1.
+ *
+ * 3. go over all drivers again, collect the quantum/rate of all followers, select
+ *    the desired final value and activate the followers and then the driver.
+ *
+ * A complete graph evaluation is performed for each change that is made to the
+ * graph, such as making/destroying links, adding/removing nodes, property changes such
+ * as quantum/rate changes or metadata changes.
+ */
 int pw_context_recalc_graph(struct pw_context *context, const char *reason)
 {
 	struct impl *impl = SPA_CONTAINER_OF(context, struct impl, this);
@@ -1089,6 +1003,7 @@ int pw_context_recalc_graph(struct pw_context *context, const char *reason)
 	uint32_t max_quantum, min_quantum, def_quantum, lim_quantum, rate_quantum;
 	uint32_t *rates, n_rates, def_rate;
 	bool freewheel = false, global_force_rate, force_rate, force_quantum, global_force_quantum;
+	struct spa_list collect;
 
 	pw_log_info("%p: busy:%d reason:%s", context, impl->recalc, reason);
 
@@ -1116,9 +1031,11 @@ again:
 		if (n->exported)
 			continue;
 
-		if (!n->visited)
-			collect_nodes(context, n);
-
+		if (!n->visited) {
+			spa_list_init(&collect);
+			collect_nodes(context, n, &collect);
+			move_to_driver(context, &collect, n);
+		}
 		/* from now on we are only interested in active driving nodes.
 		 * We're going to see if there are active followers. */
 		if (!n->driving || !n->active)
@@ -1155,29 +1072,41 @@ again:
 
 	/* now go through all available nodes. The ones we didn't visit
 	 * in collect_nodes() are not linked to any driver. We assign them
-	 * to either an active driver of the first driver */
+	 * to either an active driver or the first driver if they are in a
+	 * group that needs a driver. Else we remove them from a driver
+	 * and stop them. */
 	spa_list_for_each(n, &context->node_list, link) {
-		if (n->exported)
+		struct pw_impl_node *t, *driver;
+
+		if (n->exported || n->visited)
 			continue;
 
-		if (!n->visited) {
-			struct pw_impl_node *t;
+		pw_log_debug("%p: unassigned node %p: '%s' active:%d want_driver:%d target:%p",
+				context, n, n->name, n->active, n->want_driver, target);
 
-			pw_log_debug("%p: unassigned node %p: '%s' active:%d want_driver:%d target:%p",
-					context, n, n->name, n->active, n->want_driver, target);
+		/* collect all nodes in this group */
+		spa_list_init(&collect);
+		collect_nodes(context, n, &collect);
 
-			t = n->want_driver ? target : NULL;
-
-			pw_impl_node_set_driver(n, t);
-			if (t == NULL)
-				ensure_state(n, false);
-			else {
-				t->passive = false;
-				collect_nodes(context, n);
-			}
+		driver = NULL;
+		spa_list_for_each(t, &collect, sort_link) {
+			/* is any active and want a driver or it want process */
+			if ((t->want_driver && t->active && !n->passive) ||
+			    t->always_process)
+				driver = target;
+		}
+		if (driver != NULL) {
+			/* driver needed for this group */
+			driver->passive = false;
+			move_to_driver(context, &collect, driver);
+		} else {
+			/* no driver, make sure the nodes stops */
+			remove_from_driver(context, &collect);
 		}
-		n->visited = false;
 	}
+	/* clean up the visited flag now */
+	spa_list_for_each(n, &context->node_list, link)
+		n->visited = false;
 
 	/* assign final quantum and set state for followers and drivers */
 	spa_list_for_each(n, &context->driver_list, driver_link) {
@@ -1194,7 +1123,11 @@ again:
 		/* collect quantum and rate */
 		spa_list_for_each(s, &n->follower_list, follower_link) {
 
-			if (s->info.state > PW_NODE_STATE_SUSPENDED) {
+			if (!s->moved) {
+				/* We only try to enforce the lock flags for nodes that
+				 * are not recently moved between drivers. The nodes that
+				 * are moved should try to enforce their quantum on the
+				 * new driver. */
 				lock_quantum |= s->lock_quantum;
 				lock_rate |= s->lock_rate;
 			}
@@ -1232,6 +1165,8 @@ again:
 
 			if (s->active)
 				running = !n->passive;
+
+			s->moved = false;
 		}
 
 		if (force_quantum)
@@ -1243,9 +1178,14 @@ again:
 		if (lock_rate ||
 		    (!force_rate &&
 		    (n->info.state > PW_NODE_STATE_IDLE)))
+			/* when someone wants us to lock the rate of this driver or
+			 * when the driver is busy and we don't need to force a rate,
+			 * keep the current rate */
 			target_rate = current_rate;
 		else {
-			/* calculate desired rate */
+			/* Here we are allowed to change the rate of the driver.
+			 * Start with the default rate. If the desired rate is
+			 * allowed, switch to it */
 			target_rate = def_rate;
 			if (rate.denom != 0 && rate.num == 1) {
 				if (rates_contains(rates, n_rates, rate.denom))
@@ -1254,6 +1194,7 @@ again:
 		}
 
 		if (target_rate != current_rate) {
+			/* we doing a rate switch */
 			pw_log_info("(%s-%u) state:%s new rate:%u->%u",
 					n->name, n->info.id,
 					pw_node_state_as_string(n->info.state),
@@ -1267,6 +1208,8 @@ again:
 				if (n->info.state >= PW_NODE_STATE_IDLE)
 					suspend_driver(context, n);
 			}
+			/* we're setting the pending rate. This will become the new
+			 * current rate in the next iteration of the graph. */
 			n->current_rate = SPA_FRACTION(1, target_rate);
 			n->current_pending = true;
 			current_rate = target_rate;
@@ -1275,6 +1218,7 @@ again:
 		}
 
 		if (rate_quantum != 0 && current_rate != rate_quantum) {
+			/* the quantum values are scaled with the current rate */
 			def_quantum = def_quantum * current_rate / rate_quantum;
 			min_quantum = min_quantum * current_rate / rate_quantum;
 			max_quantum = max_quantum * current_rate / rate_quantum;
@@ -1301,11 +1245,19 @@ again:
 					n->name, n->info.id,
 					n->current_quantum,
 					quantum);
+			/* this is the new pending quantum */
 			n->current_quantum = quantum;
 			n->current_pending = true;
 		}
 
 		if (n->info.state < PW_NODE_STATE_RUNNING && n->current_pending) {
+			/* the driver node is not actually running and we have a
+			 * panding change. Apply the change to the position now so
+			 * that we have the right values when we change the node
+			 * states of the driver and followers to RUNNING below */
+			pw_log_debug("%p: apply duration:%"PRIu64" rate:%u/%u", context,
+					n->current_quantum, n->current_rate.num,
+					n->current_rate.denom);
 			n->rt.position->clock.duration = n->current_quantum;
 			n->rt.position->clock.rate = n->current_rate;
 			n->current_pending = false;
@@ -1314,6 +1266,7 @@ again:
 		pw_log_debug("%p: driving %p running:%d passive:%d quantum:%u '%s'",
 				context, n, running, n->passive, quantum, n->name);
 
+		/* first change the node states of the followers to the new target */
 		spa_list_for_each(s, &n->follower_list, follower_link) {
 			if (s == n)
 				continue;
@@ -1321,6 +1274,7 @@ again:
 					context, s, s->active, s->name);
 			ensure_state(s, running);
 		}
+		/* now that all the followers are ready, start the driver */
 		ensure_state(n, running);
 	}
 	impl->recalc = false;
@@ -1456,6 +1410,12 @@ int pw_context_set_object(struct pw_context *context, const char *type, void *va
 		}
 		entry->value = value;
 	}
+	if (spa_streq(type, SPA_TYPE_INTERFACE_ThreadUtils)) {
+		context->thread_utils = value;
+		if (context->data_loop_impl)
+			pw_data_loop_set_thread_utils(context->data_loop_impl,
+					context->thread_utils);
+	}
 	return 0;
 }
 
