@@ -167,6 +167,7 @@ do_node_remove(struct spa_loop *loop,
 		spa_loop_remove_source(loop, &this->source);
 		remove_node(this);
 	}
+	this->added = false;
 	return 0;
 }
 
@@ -210,7 +211,19 @@ static int pause_node(struct pw_impl_node *this)
 	return res;
 }
 
-static void node_activate(struct pw_impl_node *this)
+static void node_activate_outputs(struct pw_impl_node *this)
+{
+	struct pw_impl_port *port;
+
+	pw_log_debug("%p: activate", this);
+	spa_list_for_each(port, &this->output_ports, link) {
+		struct pw_impl_link *link;
+		spa_list_for_each(link, &port->links, output_link)
+			pw_impl_link_activate(link);
+	}
+}
+
+static void node_activate_inputs(struct pw_impl_node *this)
 {
 	struct pw_impl_port *port;
 
@@ -220,11 +233,6 @@ static void node_activate(struct pw_impl_node *this)
 		spa_list_for_each(link, &port->links, input_link)
 			pw_impl_link_activate(link);
 	}
-	spa_list_for_each(port, &this->output_ports, link) {
-		struct pw_impl_link *link;
-		spa_list_for_each(link, &port->links, output_link)
-			pw_impl_link_activate(link);
-	}
 }
 
 static int start_node(struct pw_impl_node *this)
@@ -232,7 +240,9 @@ static int start_node(struct pw_impl_node *this)
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 	int res = 0;
 
-	node_activate(this);
+	/* First activate the outputs so that when the node starts pushing,
+	 * we can process the outputs */
+	node_activate_outputs(this);
 
 	if (impl->pending_state >= PW_NODE_STATE_RUNNING)
 		return 0;
@@ -336,6 +346,7 @@ do_node_add(struct spa_loop *loop,
 		spa_loop_add_source(loop, &this->source);
 		add_node(this, driver);
 	}
+	this->added = true;
 	return 0;
 }
 
@@ -354,8 +365,11 @@ static void node_update_state(struct pw_impl_node *node, enum pw_node_state stat
 				error = spa_aprintf("Start error: %s", spa_strerror(res));
 			}
 		}
-		if (res >= 0)
+		if (res >= 0) {
 			pw_loop_invoke(node->data_loop, do_node_add, 1, NULL, 0, true, node);
+			/* now activate the inputs */
+			node_activate_inputs(node);
+		}
 		break;
 	default:
 		break;
@@ -592,10 +606,10 @@ static const struct pw_resource_events resource_events = {
 };
 
 static int
-global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
+global_bind(void *object, struct pw_impl_client *client, uint32_t permissions,
 	    uint32_t version, uint32_t id)
 {
-	struct pw_impl_node *this = _data;
+	struct pw_impl_node *this = object;
 	struct pw_global *global = this->global;
 	struct pw_resource *resource;
 	struct resource_data *data;
@@ -842,6 +856,7 @@ int pw_impl_node_set_driver(struct pw_impl_node *node, struct pw_impl_node *driv
 			old->name, old->info.id, driver->name, driver->info.id);
 
 	node->driver_node = driver;
+	node->moved = true;
 
 	pw_loop_invoke(node->data_loop,
 		       do_move_nodes, SPA_ID_INVALID, &driver, sizeof(struct pw_impl_node *),
@@ -1046,7 +1061,7 @@ static inline int resume_node(struct pw_impl_node *this, int status)
 		if (pw_node_activation_state_dec(state, 1)) {
 			a->status = PW_NODE_ACTIVATION_TRIGGERED;
 			a->signal_time = nsec;
-			t->signal(t->data);
+			t->signal_func(t->data);
 		}
 	}
 	return 0;
@@ -1077,6 +1092,12 @@ static inline int process_node(void *data)
 	a->status = PW_NODE_ACTIVATION_AWAKE;
 	a->awake_time = SPA_TIMESPEC_TO_NSEC(&ts);
 
+	if (!this->added) {
+		/* This should not happen here. We activate the input
+		 * links after we add the node to the graph. */
+		pw_log_warn("%p: scheduling non-active node", this);
+		return -EIO;
+	}
 	pw_log_trace_fp("%p: process %"PRIu64, this, a->awake_time);
 
 	/* when transport sync is not supported, just clear the flag */
@@ -1144,7 +1165,7 @@ static void node_on_fd_events(struct spa_source *source)
 				this->name, this->info.id, cmd - 1);
 
 		pw_log_trace_fp("%p: got process", this);
-		this->rt.target.signal(this->rt.target.data);
+		this->rt.target.signal_func(this->rt.target.data);
 	}
 }
 
@@ -1261,9 +1282,9 @@ struct pw_impl_node *pw_context_create_node(struct pw_context *context,
 	this->rt.activation = this->activation->map->ptr;
 	this->rt.target.activation = this->rt.activation;
 	this->rt.target.node = this;
-	this->rt.target.signal = process_node;
+	this->rt.target.signal_func = process_node;
 	this->rt.target.data = this;
-	this->rt.driver_target.signal = process_node;
+	this->rt.driver_target.signal_func = process_node;
 
 	reset_position(this, &this->rt.activation->position);
 	this->rt.activation->sync_timeout = DEFAULT_SYNC_TIMEOUT;
@@ -1612,7 +1633,7 @@ static int node_ready(void *data, int status)
 						state->pending, state->required);
 				dump_states(node);
 			}
-			node->rt.target.signal(node->rt.target.data);
+			node->rt.target.signal_func(node->rt.target.data);
 		}
 
 		if (node->current_pending) {
@@ -2213,7 +2234,9 @@ int pw_impl_node_set_active(struct pw_impl_node *node, bool active)
 	bool old = node->active;
 
 	if (old != active) {
-		pw_log_debug("%p: %s", node, active ? "activate" : "deactivate");
+		pw_log_debug("%p: %s registered:%d", node,
+				active ? "activate" : "deactivate",
+				node->registered);
 
 		node->active = active;
 		pw_impl_node_emit_active_changed(node, active);
@@ -2221,6 +2244,8 @@ int pw_impl_node_set_active(struct pw_impl_node *node, bool active)
 		if (node->registered)
 			pw_context_recalc_graph(node->context,
 					active ? "node activate" : "node deactivate");
+		else if (!active && node->exported)
+			pw_loop_invoke(node->data_loop, do_node_remove, 1, NULL, 0, true, node);
 	}
 	return 0;
 }
